// ***************************************************************** -*- C++ -*-
// easyaccess-test.cpp, $Rev: 1760 $
// Sample program using high-level metadata access functions

// included header files
#include <exiv2/image.hpp>
#include <exiv2/exif.hpp>
#include <exiv2/easyaccess.hpp>
#include <iostream>
#include <iomanip>
#include <cassert>

typedef Exiv2::ExifData::const_iterator (*EasyAccessFct)(const Exiv2::ExifData& ed);

struct EasyAccess {
    const char*   label_;
    EasyAccessFct findFct_;
};

static const EasyAccess easyAccess[] = {
    { "Orientation",   Exiv2::orientation  },
    { "ISO speed",     Exiv2::isoSpeed     },
    { "Flash bias",    Exiv2::flashBias    },
    { "Exposure mode", Exiv2::exposureMode },
    { "Scene mode",    Exiv2::sceneMode    },
    { "Macro mode",    Exiv2::macroMode    },
    { "Image quality", Exiv2::imageQuality },
    { "White balance", Exiv2::whiteBalance },
    { "Lens name",     Exiv2::lensName     }
};

int main(int argc, char **argv)
try {
    if (argc != 2) {
        std::cout << "Usage: " << argv[0] << " file\n";
        return 1;
    }

    Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open(argv[1]);
    assert (image.get() != 0);
    image->readMetadata();
    Exiv2::ExifData& ed = image->exifData();

    for (unsigned int i = 0; i < EXV_COUNTOF(easyAccess); ++i) {
        Exiv2::ExifData::const_iterator pos = easyAccess[i].findFct_(ed);
        if (pos != ed.end()) {
            std::cout << std::setw(13) << std::left << easyAccess[i].label_ << " ("
                      << std::setw(30) << pos->key() << ") : "
                      << pos->print(&ed) << "\n";
        }
    }

    return 0;
}
catch (Exiv2::AnyError& e) {
    std::cout << "Caught Exiv2 exception '" << e << "'\n";
    return -1;
}
